

set -ex



cd hypothesis-python
python -m pytest  tests/cover
python -m pytest  tests/datetime
python -m pytest  tests/fakefactory
python -m pytest  tests/numpy
python -m pytest  tests/pandas
python -m pytest  tests/py3
python -m pytest  tests/pytest -p pytester --runpytest subprocess
exit 0
